/*
 * gbpserial.c: communicate with a GemPC410 smart card reader
 * using GemCore protocol
 * Copyright (C) 2001,2002 Ludovic Rousseau
 * 
 * This program is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 */

/*
 * $Id: gbpserial.c,v 1.28 2002/10/15 20:03:39 rousseau Exp $
 */

#include "springcard_i.h"
#include "reader.h"
#include "pcsc_impl.h"
#include "gbp_impl.h"
#ifdef WIN32
#include "serial_win32.h"
#endif
#ifdef __linux__
#include "serial_linux.h"
#endif
#ifdef __palmos__
#include "serial_palmos.h"
#endif

#ifdef GCR_SELECT_GBP_MODE

/*
 * GBP (Gemplus Bloc Protocol) format
 *
 * +-----+-----+-----+---------+-----+
 * | NAD | PCB | LEN | DAT ... | EDC |
 * +-----+-----+-----+---------+-----+
 *
 * NAD: source and destination indentifier
 * PCB: block type
 *  I-Block (Information)
 * bit 7   6   5    4-0
 *   +---+---+---+--------+
 *   | 0 | S | 0 | unused |
 *   +---+---+---+--------+
 *   S = Sequence bit
 *
 *  R-Block (Repeat)
 * bit 7   6   5   4   3   2   1   0
 *   +---+---+---+---+---+---+---+---+
 *   | 1 | 0 | 0 | S | 0 | 0 | E | V |
 *   +---+---+---+---+---+---+---+---+
 *   V = Error verified by EDC
 *   E = Another error
 *
 *  S-Block (Synchro)
 * bit 7   6   5   4   3   2   1   0
 *   +---+---+---+---+---+---+---+---+
 *   | 1 | 1 | R | 0 | 0 | 0 | 0 | 0 |
 *   +---+---+---+---+---+---+---+---+
 *   R = Resync
 *     0 = request
 *     1 = response
 *
 * LEN: number of bytes in DAT
 * DAT: data transmitted
 * EDC: xor on NAD, PCB, LEN and DAT bytes
 *
 */

// indexes in GbpBuffer[]
#define NAD 0
#define PCB 1
#define LEN 2
#define DAT 3

// NAD values
#define NAD_HOST2IFD 0x42
#define NAD_IFD2HOST 0x24
#define NAD_XXX2HOST 0x04

// communication timeout in seconds
#define SERIAL_TIMEOUT 10

/*
 * Global variables containing the file handle of the serial port
 */
typedef struct
{
  int     channel;
  int     fd;
  char    bSeq;
}
intrFace;

WORD    ExplainGBP(const unsigned char buffer[], const LONG rv);

/*****************************************************************************
 * 
 *				WriteGBP: Send bytes to the card reader
 *
 * remark: buffer already contains the byte length in [0]
 *
 *****************************************************************************/
WORD WriteGBP(ReaderDesc * reader, DWORD length, BYTE * buffer)
{
  int     rv, len, i;
  BYTE    edc;
  BYTE    gbpbuffer[CMD_BUF_SIZE + 3];
#ifdef DEBUG_LEVEL_COMM
  char    debug_header[] = "-> 121234 ";
  sprintf(debug_header, "-> %06X ", (int) lun);
#endif

  // PDU length = length + NAD + PCB + LEN
  len = length + 3;

  // NAD byte
  gbpbuffer[NAD] = NAD_HOST2IFD;

  // PCB byte
  gbpbuffer[PCB] = (reader->bSeq << 6);

  // copy LEN and DATA
  memcpy(gbpbuffer + LEN, buffer, length);

  // checksum
  edc = 0;
  for (i = 0; i < len - 1; i++)
    edc ^= gbpbuffer[i];

  // EDC byte
  gbpbuffer[len - 1] = edc;

  // increment Seq: 0 -> 1 or 1 -> 0
  reader->bSeq = (reader->bSeq + 1) % 2;

#ifdef DEBUG_LEVEL_COMM
  DEBUG_XXD(debug_header, gbpbuffer, len);
#endif

  if (reader->fatal_error)
  {
    SCARD_Trace(TR_COMM_HIGH, "Fatal error encountered, writing %d anyway", len);
  }

  rv = WRITE_COM(reader->fd, gbpbuffer, len);

  if (reader->fatal_error)
  {
    SCARD_Trace(TR_COMM_HIGH, "Fatal error encountered, writing returned %d", rv);
  }

  if (rv < 0)
  {
    if (rv == -2)
      return GC_STATUS_HARDWARE_ERROR;
    return GC_STATUS_COMMUNICATION_ERROR;
  }

  return GC_STATUS_SUCCESS;
}                               /* WriteGBP */

/*****************************************************************************
 * 
 *				ReadGBP: Receive bytes from the card reader
 *
 * remark: buffer contains the byte length in [0]
 *
 *****************************************************************************/
WORD ReadGBP(ReaderDesc * reader, DWORD * length, BYTE * buffer)
{
  int     rv, len, to_read, already_read;
  int     i, pcb, edc, retry;
  DWORD   buffer_size;
  BYTE    gbpbuffer[CMD_BUF_SIZE + 3];
#ifdef DEBUG_LEVEL_COMM
  char    debug_header[] = "<- 121234 ";
  sprintf(debug_header, "<- %06X ", (int) lun);
#endif

  len = sizeof(gbpbuffer);
  buffer_size = *length;

  // error by default
  *length = 0;

  /************
	// Default read operation, reading of the whole buffer
  	rv = READ_COM(reader->fd, gbpbuffer, len);
	if (rv < 0)
		return GC_STATUS_UNSUCCESSFUL;
	/************/

  /************/
  retry = 1 + reader->cmd_timeout;  // JDA : TODO
  // Read header and maybe the first data byte:
  while (retry--)
  {
    rv = READ_COM(reader->fd, gbpbuffer, 4);
    if (rv != 0)
      break;
    SCARD_Trace(TR_COMM_HIGH, "ReadGBP header: read error");
  }
  reader->cmd_timeout = 0;      // JDA !!!
  //printf ("Read %d bytes: %02x %02x %02x %02x\n", rv, gbpbuffer[0], gbpbuffer[1], gbpbuffer[2], gbpbuffer[3]);
  if (rv < 0)
  {
    reader->fatal_error = TRUE;
    if (rv == -2)
      return GC_STATUS_HARDWARE_ERROR;
    return GC_STATUS_COMMUNICATION_ERROR;
  }
  /************/

  // too short
  if (rv < 4)
  {
#ifdef DEBUG_LEVEL_COMM
    DEBUG_XXD(debug_header, gbpbuffer, rv);
    ExplainGBP(gbpbuffer, rv);
#endif
    SCARD_Trace(TR_COMM_HIGH, "ReadGBP: read too short (rv=%d)", rv);
    return GC_STATUS_PROTOCOL_ERROR;
  }
  // NAD byte
  if ((gbpbuffer[NAD] != NAD_IFD2HOST) && (gbpbuffer[NAD] != NAD_XXX2HOST))
  {
#ifdef DEBUG_LEVEL_COMM
    DEBUG_XXD(debug_header, gbpbuffer, rv);
    ExplainGBP(gbpbuffer, rv);
#endif
    SCARD_Trace(TR_COMM_HIGH, "ReadGBP: wrong NAD byte");
    return GC_STATUS_NAD_UNKNOWN;
  }
  // PCB byte
  pcb = gbpbuffer[PCB];

  if (pcb & 0xA0)
  {
    WORD    ret;
#ifdef DEBUG_LEVEL_COMM
    DEBUG_XXD(debug_header, gbpbuffer, rv);
#endif
    ret = ExplainGBP(gbpbuffer, rv);
    SCARD_Trace(TR_COMM_HIGH, "ReadGBP: wrong PCB byte");
    return ret;
  }
  // LEN byte (+4 bytes of protocol)
  to_read = gbpbuffer[LEN] + 4;

  // Nb of bytes already read
  already_read = rv;

  // Nb of bytes until the end of the buffer
  len -= rv;

  while (already_read < to_read)
  {
    //rv = READ_COM(reader->fd, gbpbuffer + already_read, len);
    rv = READ_COM(reader->fd, gbpbuffer + already_read, to_read - already_read);
    if (rv <= 0)
    {
#ifdef DEBUG_LEVEL_COMM
      DEBUG_XXD(debug_header, gbpbuffer, already_read);
#endif
      SCARD_Trace(TR_COMM_HIGH, "ReadGBP block: read error");
      reader->fatal_error = TRUE;
      if (rv == -2)
        return GC_STATUS_HARDWARE_ERROR;
      return GC_STATUS_COMMUNICATION_ERROR;
    }

    already_read += rv;
    len -= rv;
  }

  // calculate checksum
  edc = 0;
  for (i = 0; i < to_read; i++)
    edc ^= gbpbuffer[i];

#ifdef DEBUG_LEVEL_COMM
  DEBUG_XXD(debug_header, gbpbuffer, to_read);
#endif

  // verify checksum
  if (edc != 0)
  {
    SCARD_Trace(TR_COMM_HIGH, "ReadGBP: wrong EDC");
    return GC_STATUS_PROTOCOL_ERROR;
  }
  // copy up to buffer_size bytes
  if (buffer_size > gbpbuffer[LEN])
    *length = gbpbuffer[LEN] + 1;
  else
    *length = buffer_size;

  memcpy(buffer, gbpbuffer + LEN, *length);

  return GC_STATUS_SUCCESS;
}                               /* ReadGBP */

/*****************************************************************************
 * 
 *				ExplainGBP: print a textual error explanation
 *
 *****************************************************************************/
WORD ExplainGBP(const unsigned char buffer[], const LONG rv)
{
  WORD    ret = 0;

  if (rv < 4)
  {
    SCARD_Trace(TR_COMM_HIGH, "GBP bloc too short");
    return GC_STATUS_PROTOCOL_ERROR;
  }

  switch (buffer[0])
  {
    case 0x42:
      SCARD_Trace(TR_COMM_HIGH, "  NAD: host to gemcore");
      break;
    case 0x24:
      SCARD_Trace(TR_COMM_HIGH, "  NAD: gemcore to host");
      break;
    default:
      SCARD_Trace(TR_COMM_HIGH, "  NAD: UNKNOWN value %02X", buffer[0]);
  }

  switch (buffer[1])
  {
    case 0x00:
      SCARD_Trace(TR_COMM_HIGH, "  PCB: I-block S=0");
      break;
    case 0x40:
      SCARD_Trace(TR_COMM_HIGH, "  PCB: I-block S=1");
      break;
    case 0x80:
      SCARD_Trace(TR_COMM_HIGH, "  PCB: R-block S=0, EDC error=0, another error=0");
      ret = GC_STATUS_PCB_REP_0;
      break;
    case 0x90:
      SCARD_Trace(TR_COMM_HIGH, "  PCB: R-block S=1, EDC error=0, another error=0");
      ret = GC_STATUS_PCB_REP_1;
      break;
    case 0x81:
      SCARD_Trace(TR_COMM_HIGH, "  PCB: R-block S=0, EDC error=1, another error=0");
      ret = GC_STATUS_PCB_REP_0;
      break;
    case 0x91:
      SCARD_Trace(TR_COMM_HIGH, "  PCB: R-block S=1, EDC error=1, another error=0");
      ret = GC_STATUS_PCB_REP_1;
      break;
    case 0x82:
      SCARD_Trace(TR_COMM_HIGH, "  PCB: R-block S=0, EDC error=0, another error=1");
      ret = GC_STATUS_PCB_REP_0;
      break;
    case 0x92:
      SCARD_Trace(TR_COMM_HIGH, "  PCB: R-block S=1, EDC error=0, another error=1");
      ret = GC_STATUS_PCB_REP_1;
      break;
    case 0x83:
      SCARD_Trace(TR_COMM_HIGH, "  PCB: R-block S=0, EDC error=1, another error=1");
      ret = GC_STATUS_PCB_REP_0;
      break;
    case 0x93:
      SCARD_Trace(TR_COMM_HIGH, "  PCB: R-block S=1, EDC error=1, another error=1");
      ret = GC_STATUS_PCB_REP_1;
      break;
    case 0xB0:
      SCARD_Trace(TR_COMM_HIGH, "  PCB: S-block resynch request");
      ret = GC_STATUS_PCB_SYNC_REQ;
      break;
    case 0xE0:
      SCARD_Trace(TR_COMM_HIGH, "  PCB: S-block resynch response");
      ret = GC_STATUS_PCB_SYNC_RESP;
      break;
    default:
      SCARD_Trace(TR_COMM_HIGH, "  PCB: UNKNOWN value");
      break;
  }

  if (rv > 4)
    switch (buffer[3])
    {
      case 0x00:
        SCARD_Trace(TR_COMM_HIGH, "  S: no error");
        break;
      case 0x01:
        SCARD_Trace(TR_COMM_HIGH, "  S: unknown driver");
        break;
      case 0x02:
        SCARD_Trace(TR_COMM_HIGH, "  S: operation impossible with this driver");
        break;
      case 0x03:
        SCARD_Trace(TR_COMM_HIGH, "  S: incorrect number of arguments");
        break;
      case 0x04:
        SCARD_Trace(TR_COMM_HIGH, "  S: reader command unknown");
        break;
      case 0x05:
        SCARD_Trace(TR_COMM_HIGH, "  S: response too long for the buffer");
        break;
      case 0x10:
        SCARD_Trace(TR_COMM_HIGH, "  S: resonse error at the card reset");
        break;
      case 0x12:
        SCARD_Trace(TR_COMM_HIGH, "  S: message too long");
        break;
      case 0x13:
        SCARD_Trace(TR_COMM_HIGH, "  S: byte reading error returned by an asynchronous cadr");
        break;
      case 0x15:
        SCARD_Trace(TR_COMM_HIGH, "  S: card powered down");
        break;
      case 0x1B:
        SCARD_Trace(TR_COMM_HIGH, "  S: a command has been sent with an incorrect number of parameters");
        break;
      case 0x1C:
        SCARD_Trace(TR_COMM_HIGH, "  S: overlap on writting flash memory");
        break;
      case 0x1D:
        SCARD_Trace(TR_COMM_HIGH, "  S: the TCK check byte is incorrect in a micropocessor card ATR");
        break;
      case 0x1E:
        SCARD_Trace(TR_COMM_HIGH, "  S: an attempt has been made to write a write-protected external memory");
        break;
      case 0x1F:
        SCARD_Trace(TR_COMM_HIGH, "  S: incorrect data has been send to external memory");
        break;
      case 0xA0:
        SCARD_Trace(TR_COMM_HIGH, "  S: error in the card reset response");
        break;
      case 0xA1:
        SCARD_Trace(TR_COMM_HIGH, "  S: card protocol error");
        break;
      case 0xA2:
        SCARD_Trace(TR_COMM_HIGH, "  S: card malfuntion");
        break;
      case 0xA3:
        SCARD_Trace(TR_COMM_HIGH, "  S: parity error");
        break;
      case 0xA4:
        SCARD_Trace(TR_COMM_HIGH, "  S: card has been chaining (T=1)");
        break;
      case 0xA5:
        SCARD_Trace(TR_COMM_HIGH, "  S: reader has aborted chaining (T=1)");
        break;
      case 0xA6:
        SCARD_Trace(TR_COMM_HIGH, "  S: resynch successfully performed by GemCore");
        break;
      case 0xA7:
        SCARD_Trace(TR_COMM_HIGH, "  S: protocol type selection (PTS) error");
        break;
      case 0xB0:
        SCARD_Trace(TR_COMM_HIGH, "  S: GemCP410 command not supported");
        break;
      case 0xCF:
        SCARD_Trace(TR_COMM_HIGH, "  S: other key already pressed");
        break;
      case 0xE4:
        SCARD_Trace(TR_COMM_HIGH, "  S: the card has just send an invalid procedure byte");
        break;
      case 0xE5:
        SCARD_Trace(TR_COMM_HIGH, "  S: the card has interrupted an exchange");
        break;
      case 0xE7:
        SCARD_Trace(TR_COMM_HIGH, "  S: error returned by the card");
        break;
      case 0xF7:
        SCARD_Trace(TR_COMM_HIGH, "  S: card removed");
        break;
      case 0xF8:
        SCARD_Trace(TR_COMM_HIGH, "  S: the cadr is consuming too much electricity");
        break;
      case 0xFB:
        SCARD_Trace(TR_COMM_HIGH, "  S: card missing");
        break;
      default:
        SCARD_Trace(TR_COMM_HIGH, "  S: UNKNOWN ERROR");
        break;
    }

  return ret;
}                               /* ExplainGBP */

BOOL SetGBPSeqNumber(ReaderDesc * reader, int seq)
{
  if (reader->bSeq != -1)
  {
    reader->bSeq = seq;
    return FALSE;
  } else
    return TRUE;

}                               /* SetGBPSeqNumber */

#endif
