/*
 * GCGBPTransport.c: send a command to a GemCore serial reader Copyright
 * (C) 2001 Ludovic Rousseau
 * 
 * This program is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 */

/*
 * $Id: GCGBPTransport.c,v 1.7 2002/09/18 19:52:44 rousseau Exp $
 *
 */

#include <string.h>

#include "springcard_i.h"
#include "pcsc_impl.h"
#include "gbp_impl.h"

#ifdef GCR_SELECT_GBP_MODE

/*
 * Transmission without errors:
 * Host              Reader
 * I(0) -----------> 
 *      <----------- I(0)
 * I(1) ----------->
 *      <----------- I(1)
 *
 * Transmission with errors
 *
 *        Case 1:
 * Host              Reader
 * I(0) -----/-----> 
 *      <----------- R(0)
 * I(0) ----------->
 *      <----------- I(0)
 *
 *        Case 2:
 * Host              Reader
 * I(0) -----------> 
 *      <----\------ I(0)
 * R(0) ----------->
 *      <----------- I(0)
 * I(1) ----------->
 *
 *        Case 3:
 * Host              Reader
 * I(0) -----/-----> 
 *      <----\------ R(0)
 * R(0) ----------->
 *      <----------- R(0)
 * I(0) ----------->
 *      <----------- I(0)
 * I(1) ----------->
 *
 *        Case 4:
 * Host              Reader
 * I(0) -----------> 
 *      <----\------ I(0)
 * R(0) -----/-----> 
 *      <----------- R(1)
 * R(0) -----------> 
 *      <----------- I(0)
 * I(1) ----------->
 * 
 *        Case 5:
 * Host              Reader
 * I(0) -----------> 
 *      <----\------ I(0)
 * R(0) -----/-----> 
 *      <----\------ R(1)
 * R(0) -----------> 
 *      <----------- I(0)
 * I(1) ----------->
 *
 *        Case 6:
 * Host              Reader
 * I(0) -----------> 
 *      <----\------ I(0)
 * R(0) -----/-----> 
 *      <----\------ R(1)
 * R(0) -----/-----> 
 *      <----------- R(1)
 * R(0) -----------> 
 *      <----------- I(0)
 * I(1) ----------->
 *
 */

#ifndef SCARD_E_COMM_DATA_LOST
#define SCARD_E_COMM_DATA_LOST SCARD_F_COMM_ERROR
#endif

/*
 * Send a command to the GBP reader
 * --------------------------------
 */
LONG GBPSendCommand(ReaderDesc * reader, DWORD nLengthIn, const BYTE pcBufferCmd[], DWORD * pnLengthOut, BYTE * pcBufferOut)
{
  BYTE    pctr_to_card_buffer[GC_TR_BUF_SIZE];
  DWORD   nlength;
  WORD    rv;
  LONG    rc;
  int     i;

  rc = SCARD_S_SUCCESS;

  if (GC_TR_BUF_SIZE <= nLengthIn)
  {
    // Buffer is too small (should not happen)
    rc = SCARD_F_INTERNAL_ERROR;
    goto end;
  }
//again:
  for (i = 0; i < MAX_TRY; i++)
  {
    memcpy(pctr_to_card_buffer + 1, pcBufferCmd, nLengthIn);
    pctr_to_card_buffer[TR_OFFSET_LNG] = (BYTE) nLengthIn;
    if ((rv = WriteGBP(reader, nLengthIn + 1, pctr_to_card_buffer)) != GC_STATUS_SUCCESS)
    {
      SCARD_Trace(TR_COMM_HIGH, "WriteGBP failed %d", rv);
      if (rv == GC_STATUS_HARDWARE_ERROR)
        rc = SCARD_E_READER_UNAVAILABLE;
      else if (rv == GC_STATUS_COMMUNICATION_ERROR)
        rc = SCARD_F_COMM_ERROR;
      else
        rc = SCARD_E_COMM_DATA_LOST;
      break;
    }

    if ((nLengthIn == 2) && (pcBufferCmd[0] == 0x0A))
    {
      // JDA : this is a change baudrate command
      i = MAX_TRY;
    }

    nlength = sizeof(pctr_to_card_buffer);
    if ((rv = ReadGBP(reader, &nlength, pctr_to_card_buffer)) != GC_STATUS_SUCCESS)
    {
      // The reader reports an error.

      // Unknown NAD, retry the command
      if (rv == GC_STATUS_NAD_UNKNOWN)
      {
        SCARD_Trace(TR_COMM_HIGH, "Wrong NAD, resent sequence");
        continue;
        //goto again;
      }
      // resent sequence 0
      if (rv == GC_STATUS_PCB_REP_0)
      {
        SCARD_Trace(TR_COMM_HIGH, "Resent sequence 0");
        SetGBPSeqNumber(reader, 0);
        continue;
        //goto again;
      }
      // resent sequence 1
      if (rv == GC_STATUS_PCB_REP_1)
      {
        SCARD_Trace(TR_COMM_HIGH, "Resent sequence 1");
        SetGBPSeqNumber(reader, 1);
        continue;
        //goto again;
      }

      if (rv == GC_STATUS_PCB_SYNC_REQ || rv == GC_STATUS_PCB_SYNC_RESP)
      {
        SCARD_Trace(TR_COMM_HIGH, "Resynch unsupported");
        rc = SCARD_E_UNSUPPORTED_FEATURE;
        break;
      }

      SCARD_Trace(TR_COMM_HIGH, "ReadGBP failed %04lX", rv);
      if (rv == GC_STATUS_HARDWARE_ERROR)
        rc = SCARD_E_READER_UNAVAILABLE;
      else if (rv == GC_STATUS_COMMUNICATION_ERROR)
        rc = SCARD_F_COMM_ERROR;
      else
        rc = SCARD_E_COMM_DATA_LOST;
      break;
    }
    if (nlength < 1)
    {
      // length byte not received
      rc = SCARD_E_COMM_DATA_LOST;
      break;
      //goto done;
    }
    nlength--;
    *pnLengthOut = (*pnLengthOut < nlength) ? *pnLengthOut : nlength;
    memcpy(pcBufferOut, pctr_to_card_buffer + 1, *pnLengthOut);
    break;
  }
//done:
end:

  if (i >= MAX_TRY)
    // Error, too much retry
    rc = SCARD_E_COMM_DATA_LOST;

  if (rc != SCARD_S_SUCCESS)
    *pnLengthOut = 0;

  // Clear buffer
  memset(pctr_to_card_buffer, 0x00, sizeof(pctr_to_card_buffer));

  return rc;
}                               /* GCSendCommand */

/*
 * Headers for APDU/ISO exchanges
 * ------------------------------
 */
const BYTE pcISOOutput[MAX_SLOT_NB] = { 0x13, 0x1B };
const BYTE pcISOInput[MAX_SLOT_NB] = { 0x14, 0x1C };
const BYTE pcEchAPDU[MAX_SLOT_NB] = { 0x15, 0x1D };
const BYTE pcLongDataADPU[] = { 0xFF, 0xFF, 0xFF, 0xFF, 0xFF };

/*
 * Status handling
 * ---------------
 */
LONG GCR_CheckStatus(DWORD nlength, DWORD * RxLength, BYTE * pcbuffer, BYTE * RxBuffer)
{
  if (nlength < GC_SIZE_STATUS)
    /* Length should at least be GC_SIZE_STATUS (reader status) */
    return SCARD_F_COMM_ERROR;

  nlength -= GC_SIZE_STATUS;
  switch (pcbuffer[GC_STATUS_OFFSET])
  {
    case GCORE_OK:
    case GCORE_NOT_9000:
    case GCORE_CARD_EXC_INT:   /* SW=6700 */
      /* if ( nlength < pcbuffer[OFFSET_LNG] ) */
      /* WHAT SHOULD WE DO                     */
      *RxLength = (*RxLength < nlength) ? (*RxLength) : nlength;
      memcpy(RxBuffer, pcbuffer + GC_OFFSET_RESP_DATA, *RxLength);
      break;

    case GCORE_CARD_MUTE:
      return SCARD_E_TIMEOUT;

    case GCORE_CARD_PROT_ERR:
      return SCARD_E_PROTO_MISMATCH;

    case GCORE_CARD_MISSING:
      return SCARD_W_REMOVED_CARD;

    default:
      // There was a problem in sending the command
      return SCARD_F_COMM_ERROR;
  }
  return SCARD_S_SUCCESS;
}                               /* GCR_CheckStatus */

LONG gemcore_long_data_INPUT_processing(ReaderDesc * reader, BYTE cCMD, DWORD ndatalength, BYTE * pcbuffer)
{
  BYTE    pccmd_buffer[CMD_BUF_SIZE];
  BYTE    pcresp_buffer[RESP_BUF_SIZE];
  BYTE    pcresp_buffer2[RESP_BUF_SIZE];
  DWORD   nlength, nlength2;
  LONG    rc;

  /* command */
  pccmd_buffer[GC_OFFSET_CMD] = cCMD;

  /* use specific APDU */
  memcpy(pccmd_buffer + GC_OFFSET_APDU, pcLongDataADPU, sizeof(pcLongDataADPU));

  /* modify the length byte */
  pccmd_buffer[GC_OFFSET_LENGTH] = (BYTE) ndatalength;

  /* Check the space left in pccmd_buffer */
  if (sizeof(pccmd_buffer) < (GC_OFFSET_TPDU_CMD_DATA + ndatalength))
  {
    rc = SCARD_E_INSUFFICIENT_BUFFER;
    goto clean_up_and_return;
  }

  /* Add the Data (last data bytes of the command) */
  memcpy(pccmd_buffer + GC_OFFSET_TPDU_CMD_DATA, pcbuffer, ndatalength);

  /* Send the command */
  nlength = sizeof(pcresp_buffer);

  rc = GBPSendCommand(reader, sizeof(pcLongDataADPU) + GC_SIZE_CMD + ndatalength, pccmd_buffer, &nlength, pcresp_buffer);
  if (rc != SCARD_S_SUCCESS)
    goto clean_up_and_return;

  nlength2 = sizeof(pcresp_buffer2);
  rc = GCR_CheckStatus(nlength, &nlength2, pcresp_buffer, pcresp_buffer2);

clean_up_and_return:
  memset(pccmd_buffer, 0x00, sizeof(pccmd_buffer));
  memset(pcresp_buffer, 0x00, sizeof(pcresp_buffer));
  memset(pcresp_buffer2, 0x00, sizeof(pcresp_buffer2));

  return rc;
}                               /* gemcore_long_data_INPUT_processing */

/*
 * nbuf_size should hold the max length available in pcbuffer 
 */
LONG gemcore_long_data_OUTPUT_processing(ReaderDesc * reader, BYTE cCMD, DWORD nbuf_size, DWORD * RxLength, BYTE * pcbuffer)
{
  BYTE    pccmd_buffer[CMD_BUF_SIZE];
  BYTE    pcresp_buffer[RESP_BUF_SIZE];
  DWORD   nlength;
  LONG    rc;

  /* command */
  pccmd_buffer[GC_OFFSET_CMD] = cCMD;

  /* use specific APDU */
  memcpy(pccmd_buffer + GC_OFFSET_APDU, pcLongDataADPU, sizeof(pcLongDataADPU));

  /* Send the command */
  nlength = sizeof(pcresp_buffer);
  rc = GBPSendCommand(reader, sizeof(pcLongDataADPU) + GC_SIZE_CMD, pccmd_buffer, &nlength, pcresp_buffer);
  if (rc != SCARD_S_SUCCESS)
    goto clean_up_and_return;

  rc = GCR_CheckStatus(nlength, &nbuf_size, pcresp_buffer, pcbuffer);
  *RxLength = *RxLength + nbuf_size;

clean_up_and_return:
  memset(pccmd_buffer, 0x00, sizeof(pccmd_buffer));
  memset(pcresp_buffer, 0x00, sizeof(pcresp_buffer));

  return rc;
}                               /* gemcore_long_data_OUTPUT_processing */

/*
 * Outgoing APDU
 * -------------
 */
LONG GBPSendOutgoingAPDU(ReaderDesc * reader, BOOL aux, BYTE * TxBuffer, DWORD TxLength, BYTE * RxBuffer, DWORD * RxLength)
{
  BYTE    pccmd_buffer[CMD_BUF_SIZE];
  BYTE    pcbuffer[RESP_BUF_SIZE];
  DWORD   nlength, ntemp_length;
  LONG    rc = SCARD_S_SUCCESS; /* Assume it will work */

  /* Buffer only holds an APDU (no Data), */
  /* output command                       */
  pccmd_buffer[GC_OFFSET_CMD] = pcISOOutput[aux];
  memcpy(pccmd_buffer + GC_OFFSET_APDU, TxBuffer, TxLength);

  /* Send the command */
  nlength = sizeof(pcbuffer);
  rc = GBPSendCommand(reader, TxLength + GC_SIZE_CMD, pccmd_buffer, &nlength, pcbuffer);
  if (rc != SCARD_S_SUCCESS)
    goto clean_up_and_return;

  // Copy RxLength to memorize the RxBuffer size
  ntemp_length = *RxLength;
  rc = GCR_CheckStatus(nlength, RxLength, pcbuffer, RxBuffer);

  /* Check if command was successful */
  if (rc != SCARD_S_SUCCESS)
    goto clean_up_and_return;

  /* Check if there is more data to fetch */
  /* First check for 0 length             */
  if (TxBuffer[ISO_OFFSET_LENGTH] == 0)
  {
    /* APDU Lenght byte is 0, check if card sent Data (256 Data) */
    if (nlength <= (ISO_SIZE_SW + GC_READER_TRANSMIT_RESP_LENGTH))
      /* Length was 0 and card returned no data: Case 1 command */
      goto clean_up_and_return;

    /* Data are left                                        */
    /* set nlength to remaining space avaliable in RxBuffer */
    nlength = ntemp_length - *RxLength;
    if (nlength <= 0)
    {
      rc = SCARD_F_COMM_ERROR;
      goto clean_up_and_return;
    }

    /* Retrieve remaining data and paste it after the first half */
    rc = gemcore_long_data_OUTPUT_processing(reader, pcISOOutput[aux], nlength, RxLength, RxBuffer + *RxLength);
  } else if ((TxBuffer[ISO_OFFSET_LENGTH] > GC_ISO_OUTPUT_MAX_DATA_LENGTH) && (nlength - ISO_SIZE_SW >= GC_ISO_OUTPUT_MAX_DATA_LENGTH))
  {
    /* Data are left                                        */
    /* set nlength to remaining space avaliable in pcbuffer */
    nlength = ntemp_length - *RxLength;

    /* Retrieve remaining data and paste it after the first half */
    rc = gemcore_long_data_OUTPUT_processing(reader, pcISOOutput[aux], nlength, RxLength, RxBuffer + *RxLength);
  }

clean_up_and_return:
  // Buffers clean-up
  memset(pccmd_buffer, 0x00, sizeof(pccmd_buffer));
  memset(pcbuffer, 0x00, sizeof(pcbuffer));

  if (rc != SCARD_S_SUCCESS)
    *RxLength = 0;

  return rc;
}                               /* gemcore_ISO_OUTPUT_processing */

/*
 * Incoming APDU
 * -------------
 */
LONG GBPSendIncomingAPDU(ReaderDesc * reader, BOOL aux, BYTE * TxBuffer, DWORD TxLength, BYTE * RxBuffer, DWORD * RxLength)
{
  BYTE    pccmd_buffer[CMD_BUF_SIZE];
  BYTE    pcbuffer[RESP_BUF_SIZE];
  DWORD   nlength;
  LONG    rc = SCARD_S_SUCCESS; // Assume it will work

  nlength = TxBuffer[ISO_OFFSET_LENGTH];
#if 0
  /*
   * this part of code is disabled since a nlength == 0 si used for Case1
   * APDUs (CLA INS P1 P2). They are tranfsormed in ifdhandler.c in
   * ISO_INPUT (CLA INS P1 P2 00) TPDUs
   */

  if (!nlength)
  {
    /* Data field present but Lc=0, forbidden */
    rc = SCARD_F_COMM_ERROR;
    goto clean_up_and_return;
  }
#endif

  if (TxLength < (nlength + ISO_CMD_SIZE + ISO_LENGTH_SIZE))
  {
    /* Not enough Data in TxBuffer */
    rc = SCARD_F_COMM_ERROR;
    goto clean_up_and_return;
  }

  /* If length exceeds maximum for one exchange, send the last part */
  /* first                                                          */
  if (nlength > GC_ISO_INPUT_MAX_DATA_LENGTH)
  {
    rc =
      gemcore_long_data_INPUT_processing(reader, pcISOInput[aux], nlength - GC_ISO_INPUT_MAX_DATA_LENGTH,
                                         TxBuffer + ISO_CMD_SIZE + ISO_LENGTH_SIZE + GC_ISO_INPUT_MAX_DATA_LENGTH);

    if (rc != SCARD_S_SUCCESS)
      goto clean_up_and_return;

    /* Length to send is now smaller */
    TxLength -= nlength - GC_ISO_INPUT_MAX_DATA_LENGTH;
  }

  nlength = sizeof(pcbuffer);
  pccmd_buffer[GC_OFFSET_CMD] = pcISOInput[aux];
  memcpy(pccmd_buffer + GC_OFFSET_APDU, TxBuffer, TxLength);

  rc = GBPSendCommand(reader, TxLength + GC_SIZE_CMD, pccmd_buffer, &nlength, pcbuffer);
  if (rc != SCARD_S_SUCCESS)
    goto clean_up_and_return;

  rc = GCR_CheckStatus(nlength, RxLength, pcbuffer, RxBuffer);

clean_up_and_return:
  /* Buffers clean-up */
  memset(pccmd_buffer, 0x00, sizeof(pccmd_buffer));
  memset(pcbuffer, 0x00, sizeof(pcbuffer));

  if (rc != SCARD_S_SUCCESS)
    *RxLength = 0;

  return rc;
}                               /* gemcore_ISO_INPUT_processing */

/*
 * Exchange APDU (T=1)
 * -------------------
 */
LONG GBPSendExchangeAPDU(ReaderDesc * reader, BOOL aux, BYTE * TxBuffer, DWORD TxLength, BYTE * RxBuffer, DWORD * RxLength)
{
  BYTE    pccmd_buffer[CMD_BUF_SIZE];
  BYTE    pcbuffer[RESP_BUF_SIZE];
  DWORD   nlength;
  LONG    rc = SCARD_S_SUCCESS; /* Assume it will work */

  /* If length exceeds maximum for one exchange, send the last part */
  /* first                                                          */
  if (TxLength > GC_ISO_EXC_CMD_MAX_APDU_LENGTH)
  {
    rc = gemcore_long_data_INPUT_processing(reader, pcEchAPDU[aux], TxLength - GC_ISO_EXC_CMD_MAX_APDU_LENGTH, TxBuffer + GC_ISO_EXC_CMD_MAX_APDU_LENGTH);

    if (rc != SCARD_S_SUCCESS)
      goto clean_up_and_return;

    /* Length to send is now smaller */
    TxLength = GC_ISO_EXC_CMD_MAX_APDU_LENGTH;
  }

  /* Build the commmand */
  pccmd_buffer[GC_OFFSET_CMD] = pcEchAPDU[aux];
  memcpy(pccmd_buffer + GC_OFFSET_APDU, TxBuffer, TxLength);

  /* Add length of expected data */
  pccmd_buffer[GC_OFFSET_APDU + TxLength] = (BYTE) * RxLength;

  /* Send the command */
  nlength = sizeof(pcbuffer);
  rc = GBPSendCommand(reader, TxLength + GC_SIZE_CMD, pccmd_buffer, &nlength, pcbuffer);
  if (rc != SCARD_S_SUCCESS)
    goto clean_up_and_return;

  if (nlength < GC_SIZE_STATUS)
  {
    /* Length should at least be GC_SIZE_STATUS (reader status) */
    return SCARD_F_COMM_ERROR;
  }

  if (pcbuffer[GC_STATUS_OFFSET] == GCORE_MORE_DATA)
  {

    if (nlength != (GC_ISO_EXC_RESP_MAX_APDU_LENGTH + GC_SIZE_STATUS))
    {
      /* Should have received a full block */
      rc = SCARD_F_COMM_ERROR;
      goto clean_up_and_return;
    }

    /* There are extra data to fetch on the reader */
    /* store what was receive                      */
    if (*RxLength <= (GC_ISO_EXC_RESP_MAX_APDU_LENGTH))
    {
      memcpy(RxBuffer, pcbuffer + GC_OFFSET_RESP_DATA, *RxLength);
      /* No space left in buffer, return directly */
      goto clean_up_and_return;
    }

    memcpy(RxBuffer, pcbuffer + GC_OFFSET_RESP_DATA, GC_ISO_EXC_RESP_MAX_APDU_LENGTH);

    /* set nlength to remaining space avaliable in RxBuffer */
    nlength = *RxLength - GC_ISO_EXC_RESP_MAX_APDU_LENGTH;

    /* set RxLength to what was already received */
    *RxLength = GC_ISO_EXC_RESP_MAX_APDU_LENGTH;

    /* Retrieve remaining data and paste it after the first half */
    rc = gemcore_long_data_OUTPUT_processing(reader, pcEchAPDU[aux], nlength, RxLength, RxBuffer + *RxLength);
  } else
    rc = GCR_CheckStatus(nlength, RxLength, pcbuffer, RxBuffer);

clean_up_and_return:
  /* Buffers clean-up */
  memset(pccmd_buffer, 0x00, sizeof(pccmd_buffer));
  memset(pcbuffer, 0x00, sizeof(pcbuffer));

  if (rc != SCARD_S_SUCCESS)
    *RxLength = 0;

  return rc;
}                               /* gemcore_ISO_EXCHANGE_processing */

LONG GBPSendApdu(ReaderDesc * reader, BOOL aux, DWORD protocol, BYTE * TxBuffer, DWORD TxLength, BYTE * RxBuffer, DWORD * RxLength)
{
  DWORD   ntestlength;
  LONG    return_value = SCARD_S_SUCCESS; // Assume it will work

  switch (protocol)
  {
    case SCARD_PROTOCOL_T0:
      // Check if command is going to fit in buffer
      if (CMD_BUF_SIZE < (GC_SIZE_CMD + TxLength))
      {
        // Buffer too small, send an error
        return_value = SCARD_E_INSUFFICIENT_BUFFER;
        goto clean_up_and_return;
      }
      // Check if this is an incoming or outgoing command
      // Size should be command + one byte of length for
      // an outgoing TPDU (CLA, INS, P1, P2, P3)
      if (TxLength == (ISO_CMD_SIZE + ISO_LENGTH_SIZE))
      {
        return_value = GBPSendOutgoingAPDU(reader, aux, TxBuffer, TxLength, RxBuffer, RxLength);
        break;
      } else
      {
        // just (CLA, INS, P1, P2) for an APDU
        if (TxLength == ISO_CMD_SIZE)
        {
          char    cmd[ISO_CMD_SIZE + ISO_LENGTH_SIZE];

          // copy CLA, INS, P1, P2
          memcpy(cmd, TxBuffer, ISO_CMD_SIZE);
          // set P3 to 0
          cmd[ISO_CMD_SIZE] = 0;
          return_value = GBPSendIncomingAPDU(reader, aux, cmd, ISO_CMD_SIZE + ISO_LENGTH_SIZE, RxBuffer, RxLength);
        } else
        {
          if (TxLength > (ISO_CMD_SIZE + ISO_LENGTH_SIZE))
          {
            // Check length to see if it is a full APDU or a TPDU
            ntestlength = TxBuffer[ISO_OFFSET_LENGTH] + ISO_CMD_SIZE + ISO_LENGTH_SIZE;

            if (TxLength == (ntestlength + ISO_LENGTH_SIZE))
            {
              // TxBuffer holds a proper APDU
              return_value = GBPSendExchangeAPDU(reader, aux, TxBuffer, TxLength, RxBuffer, RxLength);
              break;
            } else if (TxLength > (ntestlength + ISO_LENGTH_SIZE))
            {
              // Data are too long
              return_value = SCARD_E_INSUFFICIENT_BUFFER;

              goto clean_up_and_return;
            }
            // Incoming TPDU
            return_value = GBPSendIncomingAPDU(reader, aux, TxBuffer, TxLength, RxBuffer, RxLength);
          } else
          {
            // TxBuffer holds too little data to form an APDU+length
            return_value = SCARD_E_COMM_DATA_LOST;
            goto clean_up_and_return;
          }
        }
      }

      break;

    case SCARD_PROTOCOL_T1:
      // Check if command is going to fit in buffer
      // cmd byte + TxLength + Le
      if (CMD_BUF_SIZE < (GC_SIZE_CMD + TxLength + ISO_LENGTH_SIZE))
      {
        // Buffer too small, send an error
        return_value = SCARD_E_INSUFFICIENT_BUFFER;
        goto clean_up_and_return;
      }

      return_value = GBPSendExchangeAPDU(reader, aux, TxBuffer, TxLength, RxBuffer, RxLength);

      break;

    default:
      return_value = SCARD_E_PROTO_MISMATCH;
  }

clean_up_and_return:
  if (return_value != SCARD_S_SUCCESS)
    *RxLength = 0;

  return return_value;

}
#endif
