/**h* SpringCardS10API/S10
 *
 * NAME
 *   SpringCardS10API :: S10 Card entry functions
 *
 * COPYRIGHT
 *   (c) 2009 SpringCard - www.springcard.com
 *
 * DESCRIPTION
 *   Implementation of S10 entry functions to read or write the card
 *
 **/

#include "springcard_s10_i.h"

/**f* SpringCardS10API/SCardS10ReadDataArea
 *
 * NAME
 *   SCardS10ReadDataArea
 *
 * DESCRIPTION
 *   S10 read main memory function
 *
 * INPUTS
 *   SCARDHANDLE hCard : handle of the card
 *   BYTE *recv_data   : data received
 *   BYTE addr         : first address to read
 *   BYTE recv_len     : length of data to read in byte
 *   BYTE status[2]    : 2-byte buffer to receive status code
 *
 * RETURNS
 *   SCARD_S_SUCCESS   : success, card read
 *   Other code if internal or communication error has occured.  
 *
 * NOTES
 *   Maximun length of data to read is 128 bytes
 *
 * SEE ALSO
 *   SCardS10ReadProtectionArea
 *   SCardS10ReadSecurityArea
 *
 **/
SPRINGCARD_S10_LIB LONG SPRINGCARD_S10_API SCardS10ReadDataArea(SCARDHANDLE hCard, BYTE *recv_data, BYTE addr, BYTE recv_len, BYTE status[2])
{ 
  LONG SCardLastError;

  switch (SCardGetGCRFamily())
  {
    case GCR_POS      : SCardLastError = PosS10ReadDataArea(hCard, recv_data, addr, recv_len, status);
						break;

	  case GCR_ORIGINAL :
    case GCR_LITE     :
    case GCR_LITE_PRO :
	  default           : SCardLastError = LiteProS10ReadDataArea(hCard, recv_data, addr, recv_len, status);
						break;
  }

  return SCardLastError;
}

/**f* SpringCardS10API/SCardS10ReadProtectionArea
 *
 * NAME
 *   SCardS10ReadProtectionArea
 *
 * DESCRIPTION
 *   S10 read protection memory function
 *
 * INPUTS
 *   SCARDHANDLE hCard : handle of the card
 *   BYTE recv_data[4] : 4-byte buffer to receive protection data
 *   BYTE status[2]    : 2-byte buffer to receive status code
 *
 * RETURNS
 *   SCARD_S_SUCCESS    : success, protection memory read
 *   Other code if internal or communication error has occured.  
 *
 * NOTES
 *   This function must be used with a S10 card containing a protection memory.
 *   For more information, read the datasheet of the S10 card.
 *
 * SEE ALSO
 *   SCardS10ReadDataArea
 *   SCardS10ReadSecurityArea
 *
 **/
SPRINGCARD_S10_LIB LONG SPRINGCARD_S10_API SCardS10ReadProtectionArea(SCARDHANDLE hCard, BYTE recv_data[4], BYTE status[2])
{ 
  LONG SCardLastError;

  switch (SCardGetGCRFamily())
  {
    case GCR_POS      : SCardLastError = PosS10ReadProtectionArea(hCard, recv_data, status);
						break;

	  case GCR_ORIGINAL :
    case GCR_LITE     :
    case GCR_LITE_PRO :
	  default           : SCardLastError = LiteProS10ReadProtectionArea(hCard, recv_data, status);
						break;
  }

  return SCardLastError;
}

/**f* SpringCardS10API/SCardS10ReadSecurityArea
 *
 * NAME
 *   SCardS10ReadSecurityArea
 *
 * DESCRIPTION
 *   S10 read security memory function
 *
 * INPUTS
 *   SCARDHANDLE hCard : handle of the card
 *   BYTE recv_data[4] : 4-byte buffer to receive security data
 *   BYTE status[2]    : 2-byte buffer to receive status code
 *
 * RETURNS
 *   SCARD_S_SUCCESS    : success, security memory read
 *   Other code if internal or communication error has occured.  
 *
 * NOTES
 *   This function must be used with a S10 card containing a security memory.
 *   For more information, read the datasheet of the S10 card.
 * 
 * SEE ALSO
 *   SCardS10ReadDataArea
 *   SCardS10ReadProtectionArea
 *
 **/
SPRINGCARD_S10_LIB LONG SPRINGCARD_S10_API SCardS10ReadSecurityArea(SCARDHANDLE hCard, BYTE recv_data[4], BYTE status[2])
{ 
  LONG SCardLastError;

  switch (SCardGetGCRFamily())
  {
    case GCR_POS      : SCardLastError = PosS10ReadSecurityArea(hCard, recv_data, status);
						break;

	  case GCR_ORIGINAL :
    case GCR_LITE     :
    case GCR_LITE_PRO :
	  default           : SCardLastError = LiteProS10ReadSecurityArea(hCard, recv_data, status);
						break;
  }

  return SCardLastError;
}

/**f* SpringCardS10API/SCardS10WriteDataArea
 *
 * NAME
 *   SCardS10WriteDataArea
 *
 * DESCRIPTION
 *   S10 write data memory function
 *
 * INPUTS
 *   SCARDHANDLE hCard : handle of the card
 *   BYTE *write_data  : data to write
 *   BYTE addr         : first address to write
 *   BYTE write_len    : length of data to write
 *   BYTE status[2]    : 2-byte buffer to receive status code
 *
 * RETURNS
 *   SCARD_S_SUCCESS   : success, data written
 *   Other code if internal or communication error has occured.  
 *
 * NOTES
 *   Maximun length of data to write is 128 bytes.
 *   If the S10 card has a PSC, this function will only be executed if a PSC has been successfully verified before.
 *   For more information, read the datasheet of the S10 card.
 * 
 * SEE ALSO
 *   SCardS10WriteProtectionArea
 *   SCardS10UpdateSecurityArea
 *
 **/
SPRINGCARD_S10_LIB LONG SPRINGCARD_S10_API SCardS10WriteDataArea(SCARDHANDLE hCard, BYTE *write_data, BYTE addr, BYTE write_len, BYTE status[2])
{ 
  LONG SCardLastError;

  switch (SCardGetGCRFamily())
  {
    case GCR_POS      : SCardLastError = PosS10WriteDataArea(hCard, write_data, addr, write_len, status);
						break;

	  case GCR_ORIGINAL :
    case GCR_LITE     :
    case GCR_LITE_PRO :
	  default           : SCardLastError = LiteProS10WriteDataArea(hCard, write_data, addr, write_len, status);
						break;
  }

  return SCardLastError;
}

/**f* SpringCardS10API/SCardS10WriteProtectionArea
 *
 * NAME
 *   SCardS10WriteProtectionArea
 *
 * DESCRIPTION
 *   S10 write protection memory function
 *
 * WARNING
 *   The protection bytes are irreversible byte-wise.
 *
 * INPUTS
 *   SCARDHANDLE hCard : handle of the card
 *   BYTE *write_data  : data to write
 *   BYTE addr         : first address to write
 *   BYTE write_len    : length of data to write
 *   BYTE status[2]    : 2-byte buffer to receive status code
 *
 * RETURNS
 *   SCARD_S_SUCCESS   : success, data written
 *   Other code if internal or communication error has occured.  
 *
 * NOTES
 *   This function must be used with a S10 card containing a protection memory.
 *   This function will only be executed if a PSC has been successfully verified before.
 *   Maximum length of data to write is 32 bytes.
 *   Address lies between 0 and 31.
 *   For more information, read the datasheet of the S10 card.
 * 
 * SEE ALSO
 *   SCardS10WriteDataArea
 *   SCardS10UpdateSecurityArea
 *
 **/
SPRINGCARD_S10_LIB LONG SPRINGCARD_S10_API SCardS10WriteProtectionArea(SCARDHANDLE hCard, BYTE *write_data, BYTE addr, BYTE write_len, BYTE status[2])
{ 
  LONG SCardLastError;

  switch (SCardGetGCRFamily())
  {
    case GCR_POS      : SCardLastError = PosS10WriteProtectionArea(hCard, write_data, addr, write_len, status);
						break;

	  case GCR_ORIGINAL :
    case GCR_LITE     :
    case GCR_LITE_PRO :
	  default           : SCardLastError = LiteProS10WriteProtectionArea(hCard, write_data, addr, write_len, status);
						break;
  }

  return SCardLastError;
}

/**f* SpringCardS10API/SCardS10UpdateSecurityArea
 *
 * NAME
 *   SCardS10UpdateSecurityArea
 *
 * DESCRIPTION
 *   S10 update security memory function
 *
 * INPUTS
 *   SCARDHANDLE hCard : handle of the card
 *   BYTE *update_data : data to update
 *   BYTE addr         : first address to update
 *   BYTE update_len   : length of data to update
 *   BYTE status[2]    : 2-byte buffer to receive status code
 *
 * RETURNS
 *   SCARD_S_SUCCESS   : success, data updated
 *   Other code if internal or communication error has occured.  
 *
 * NOTES
 *   This function must be used with a S10 card containing a security memory.
 *   This function will only be executed if a PSC has been successfully verified before.
 *   Maximum length of data to write is 4 bytes.
 *   Address lies between 0 and 3.
 *   For more information, read the datasheet of the S10 card.
 * 
 * SEE ALSO
 *   SCardS10WriteDataArea
 *   SCardS10WriteProtectionArea
 *
 **/
SPRINGCARD_S10_LIB LONG SPRINGCARD_S10_API SCardS10UpdateSecurityArea(SCARDHANDLE hCard, BYTE *update_data, BYTE addr, BYTE update_len, BYTE status[2])
{ 
  LONG SCardLastError;

  switch (SCardGetGCRFamily())
  {
    case GCR_POS      : SCardLastError = PosS10UpdateSecurityArea(hCard, update_data, addr, update_len, status);
						break;

	  case GCR_ORIGINAL :
    case GCR_LITE     :
    case GCR_LITE_PRO :
	  default           : SCardLastError = LiteProS10UpdateSecurityArea(hCard, update_data, addr, update_len, status);
						break;
  }

  return SCardLastError;
}

/**f* SpringCardS10API/SCardS10PSCVerification
 *
 * NAME
 *   SCardS10PSCVerification
 *
 * DESCRIPTION
 *   S10 verification function of the Programmable Security Code (PSC) 
 *
 * INPUTS
 *   SCARDHANDLE hCard  : handle of the card
 *   BYTE code3         : byte 3 of Programmable Security Code
 *   BYTE code2         : byte 2 of Programmable Security Code
 *   BYTE code1         : byte 1 of Programmable Security Code
 *   BYTE status[2]     : 2-byte buffer to receive status code
 *
 * RETURNS
 *   SCARD_S_SUCCESS     : PSC verification succeeded
 *   Other code if internal or communication error has occured.  
 *
 * NOTES
 *   This function must be used with a S10 card containing a Programmable Security Code.
 *   For more information, read the datasheet of the S10 card.
 * 
 **/
SPRINGCARD_S10_LIB LONG SPRINGCARD_S10_API SCardS10PSCVerification(SCARDHANDLE hCard, BYTE code3, BYTE code2, BYTE code1, BYTE status[2])
{ 
  LONG SCardLastError;

  switch (SCardGetGCRFamily())
  {
    case GCR_POS      : SCardLastError = PosS10PSCVerification(hCard, code3, code2, code1, status);
						break;

	  case GCR_ORIGINAL :
    case GCR_LITE     :
    case GCR_LITE_PRO :
	  default           : SCardLastError = LiteProS10PSCVerification(hCard, code3, code2, code1, status);
						break;
  }

  return SCardLastError;
}
