/* SpringCardS10API/S10_GCR_1.5X
 *
 * NAME
 *   SpringCardS10API :: S10 Card specific functions for GemCore POS Serial V1.5X
 *
 * COPYRIGHT
 *   (c) 2009 SpringCard - www.springcard.com
 *
 * DESCRIPTION
 *   Implementation of S10 functions to read or write the card
 *   For GemCore POS Serial V1.5X
 *
 **/

#include "springcard_s10_i.h"

/* SpringCardS10API/POS_S10_READ_DATA_AREA_CMD
 *
 * NAME
 *   POS_S10_READ_DATA_AREA_CMD
 *
 * DESCRIPTION
 *   S10 read data APDU command
 *   For GemCore POS Serial V1.5X
 *
 **/
static BYTE POS_S10_READ_DATA_AREA_CMD[] =
{
  0x00, 0xB0, 0x00, 0x00 /* Address */, 0x00 /* Read length*/ 	
};

/* SpringCardS10API/PosS10ReadDataArea
 *
 * NAME
 *   PosS10ReadDataArea
 *
 * DESCRIPTION
 *   S10 read main memory function using APDU exchange
 *   For GemCore POS Serial V1.5X
 *
 * INPUTS
 *   SCARDHANDLE hCard : handle of the card
 *   BYTE *recv_data   : data received
 *   BYTE addr         : first address to read
 *   BYTE recv_len     : length of data to receive
 *   BYTE status[2]    : 2-byte buffer to receive status code
 *
 * RETURNS
 *   SCARD_S_SUCCESS   : success, card read
 *   Other code if internal or communication error has occured.  
 *
 * NOTES
 *   Maximun length of data to read is 128 bytes
 *
 * SEE ALSO
 *   LiteProS10ReadDataArea
 *   ScardS10ReadDataArea
 *
 **/
LONG PosS10ReadDataArea(SCARDHANDLE hCard, BYTE *recv_data, BYTE addr, BYTE recv_len, BYTE status[2])
{ 
  LONG SCardLastError;
  BYTE buffer[MAX_BYTE_TO_READ + 2];

  if (recv_len > MAX_BYTE_TO_READ) 
    return SCARD_E_INVALID_PARAMETER;

  if (recv_data == NULL) 
    return SCARD_E_INVALID_PARAMETER;

  POS_S10_READ_DATA_AREA_CMD[3] = addr;
  POS_S10_READ_DATA_AREA_CMD[4] = recv_len;

  SCardLastError = SCardS10CardTransmit(hCard, sizeof(POS_S10_READ_DATA_AREA_CMD), POS_S10_READ_DATA_AREA_CMD, sizeof(buffer), buffer, status);
  if (SCardLastError != SCARD_S_SUCCESS)
    return SCardLastError; 

  memcpy(recv_data, buffer, recv_len);

  return SCARD_S_SUCCESS;
}

/* SpringCardS10API/POS_S10_READ_PROTECTION_AREA_CMD
 *
 * NAME
 *   POS_S10_READ_PROTECTION_AREA_CMD
 *
 * DESCRIPTION
 *   S10 read protection APDU command
 *   For GemCore POS Serial V1.5X
 *
 **/
static BYTE POS_S10_READ_PROTECTION_AREA_CMD[] =
{
  0x00, 0xB0, 0x80, 0x00, 0x04	
};

/* SpringCardS10API/PosS10ReadProtectionArea
 *
 * NAME
 *   PosS10ReadProtectionArea
 *
 * DESCRIPTION
 *   S10 read protection memory function using APDU exchange
 *   For GemCore POS Serial V1.5X
 *
 * INPUTS
 *   SCARDHANDLE hCard : handle of the card
 *   BYTE recv_data[4] : 4-byte buffer to receive protection data
 *   BYTE status[2]    : 2-byte buffer to receive status code
 *
 * RETURNS
 *   SCARD_S_SUCCESS   : success, protection memory read
 *   Other code if internal or communication error has occured.  
 *
 * NOTES
 *   This function must be used with a S10 card containing a protection memory.
 *   For more information, read the datasheet of the S10 card.
 * 
 * SEE ALSO
 *   LiteProS10ReadProtectionArea
 *   SCardS10ReadProtectionArea
 *
 **/
LONG PosS10ReadProtectionArea(SCARDHANDLE hCard, BYTE recv_data[4], BYTE status[2])
{ 
  LONG SCardLastError;
  BYTE buffer[MAX_BYTE_TO_READ + 2];

  if (recv_data == NULL) 
    return SCARD_E_INVALID_PARAMETER;

  SCardLastError = SCardS10CardTransmit(hCard, sizeof(POS_S10_READ_PROTECTION_AREA_CMD), POS_S10_READ_PROTECTION_AREA_CMD, sizeof(buffer), buffer, status);
  if (SCardLastError != SCARD_S_SUCCESS)
    return SCardLastError; 

  memcpy(recv_data, buffer, 4);

  return SCARD_S_SUCCESS;
}

/* SpringCardS10API/POS_S10_READ_SECURITY_AREA_CMD
 *
 * NAME
 *   POS_S10_READ_SECURITY_AREA_CMD
 *
 * DESCRIPTION
 *   S10 read security APDU command
 *   For GemCore POS Serial V1.5X
 *
 **/
static BYTE POS_S10_READ_SECURITY_AREA_CMD[] =
{
  0x00, 0xB0, 0xC0, 0x00, 0x04	
};

/* SpringCardS10API/PosS10ReadSecurityArea
 *
 * NAME
 *   PosS10ReadSecurityArea
 *
 * DESCRIPTION
 *   S10 read security memory function using APDU exchange
 *   For GemCore POS Serial V1.5X
 *
 * INPUTS
 *   SCARDHANDLE hCard : handle of the card
 *   BYTE recv_data[4] : 4-byte buffer to receive security data
 *   BYTE status[2]    : 2-byte buffer to receive status code
 *
 * RETURNS
 *   SCARD_S_SUCCESS   : success, security memory read
 *   Other code if internal or communication error has occured.  
 *
 * NOTES
 *   This function must be used with a S10 card containing a security memory.
 *   For more information, read the datasheet of the S10 card.
 * 
 * SEE ALSO
 *   LiteProS10ReadSecurityArea
 *   SCardS10ReadSecurityArea
 *
 **/
LONG PosS10ReadSecurityArea(SCARDHANDLE hCard, BYTE recv_data[4], BYTE status[2])
{ 
  LONG SCardLastError;
  BYTE buffer[MAX_BYTE_TO_READ + 2];

  if (recv_data == NULL) 
    return SCARD_E_INVALID_PARAMETER;

  SCardLastError = SCardS10CardTransmit(hCard, sizeof(POS_S10_READ_SECURITY_AREA_CMD), POS_S10_READ_SECURITY_AREA_CMD, sizeof(buffer), buffer, status);
  if (SCardLastError != SCARD_S_SUCCESS)
    return SCardLastError; 

  memcpy(recv_data, buffer, 4);

  return SCARD_S_SUCCESS;
}

/* SpringCardS10API/POS_S10_WRITE_DATA_AREA_CMD
 *
 * NAME
 *   POS_S10_WRITE_DATA_AREA_CMD
 *
 * DESCRIPTION
 *   S10 write data APDU command
 *   For GemCore POS Serial V1.5X
 *
 **/
static BYTE POS_S10_WRITE_DATA_AREA_CMD[] =
{
  0x00, 0xD0, 0x00, 0x00 /* Address */, 0x00 /* Write length*/ 	
};

/* SpringCardS10API/PosS10WriteDataArea
 *
 * NAME
 *   PosS10WriteDataArea
 *
 * DESCRIPTION
 *   S10 write data memory function using APDU exchange
 *   For GemCore POS Serial V1.5X
 *
 * INPUTS
 *   SCARDHANDLE hCard : handle of the card
 *   BYTE *write_data  : data to write
 *   BYTE addr         : first address to write
 *   BYTE write_len    : length of data to write
 *   BYTE status[2]    : 2-byte buffer to receive status code
 *
 * RETURNS
 *   SCARD_S_SUCCESS   : success, data written
 *   Other code if internal or communication error has occured.  
 *
 * NOTES
 *   Maximun length of data to write is 128 bytes.
 *   If the S10 card has a PSC, this function will only be executed if a PSC has been successfully verified before.
 *   For more information, read the datasheet of the S10 card.
 * 
 * SEE ALSO
 *   LiteProS10WriteDataArea
 *   SCardS10WriteDataArea
 *
 **/
LONG PosS10WriteDataArea(SCARDHANDLE hCard, BYTE *write_data, BYTE addr, BYTE write_len, BYTE status[2])
{ 
  int i;
  BYTE buffer_len = 0;
  BYTE buffer[MAX_BYTE_TO_READ + 7];

  if (write_len > MAX_BYTE_TO_READ) 
    return SCARD_E_INVALID_PARAMETER;

  if (write_data == NULL) 
    return SCARD_E_INVALID_PARAMETER;

  POS_S10_WRITE_DATA_AREA_CMD[3] = addr;
  POS_S10_WRITE_DATA_AREA_CMD[4] = write_len;

  for (i=0 ; i < sizeof(POS_S10_WRITE_DATA_AREA_CMD) ; i++)
	  buffer[i] = POS_S10_WRITE_DATA_AREA_CMD[i];
  for (i=0 ; i < write_len ; i++)
	  buffer[sizeof(POS_S10_WRITE_DATA_AREA_CMD) + i] = write_data[i];

  buffer_len = sizeof(POS_S10_WRITE_DATA_AREA_CMD) + write_len;
  if (!SCardS10CardTransmit(hCard, buffer_len, buffer, sizeof(buffer), buffer, status))
      return FALSE;

  return TRUE;
}

/* SpringCardS10API/POS_S10_WRITE_PROTECTION_AREA_CMD
 *
 * NAME
 *   POS_S10_WRITE_PROTECTION_AREA_CMD
 *
 * DESCRIPTION
 *   S10 write protection APDU command
 *   For GemCore POS Serial V1.5X
 *
 **/
static BYTE POS_S10_WRITE_PROTECTION_AREA_CMD[] =
{
  0x00, 0xD0, 0x80, 0x00 /* Address */, 0x00 /* Write length*/ 	
};


/* SpringCardS10API/PosS10WriteProtectionArea
 *
 * NAME
 *   PosS10WriteProtectionArea
 *
 * DESCRIPTION
 *   S10 write protection memory function using APDU exchange
 *   For GemCore POS Serial V1.5X
 *
 * WARNING
 *   The protection bytes are irreversible byte-wise.
 *
 * INPUTS
 *   SCARDHANDLE hCard : handle of the card
 *   BYTE *write_data  : data to write
 *   BYTE addr         : first address to write
 *   BYTE write_len    : length of data to write
 *   BYTE status[2]    : 2-byte buffer to receive status code
 *
 * RETURNS
 *   SCARD_S_SUCCESS   : success, data written
 *   Other code if internal or communication error has occured.  
 *
 * NOTES
 *   This function must be used with a S10 card containing a protection memory.
 *   This function will only be executed if a PSC has been successfully verified before.
 *   Maximum length of data to write is 32 bytes.
 *   Address lies between 0 and 31.
 *   For more information, read the datasheet of the S10 card.
 * 
 * SEE ALSO
 *   LiteProS10WriteProtectionArea
 *   SCardS10WriteProtectionArea
 *
 **/
LONG PosS10WriteProtectionArea(SCARDHANDLE hCard, BYTE *write_data, BYTE addr, BYTE write_len, BYTE status[2])
{ 
  int i;
  BYTE buffer_len = 0;
  BYTE buffer[MAX_BYTE_TO_READ + 7];

  if (write_len > 32) 
    return SCARD_E_INVALID_PARAMETER;

  if (write_data == NULL) 
    return SCARD_E_INVALID_PARAMETER;

  if (! ((addr >= 0x00) && (addr <= 0x1F)))
    return SCARD_E_INVALID_PARAMETER;

  POS_S10_WRITE_PROTECTION_AREA_CMD[3] = addr;
  POS_S10_WRITE_PROTECTION_AREA_CMD[4] = write_len;

  for (i=0 ; i < sizeof(POS_S10_WRITE_PROTECTION_AREA_CMD) ; i++)
	  buffer[i] = POS_S10_WRITE_PROTECTION_AREA_CMD[i];
  for (i=0 ; i < write_len ; i++)
	  buffer[sizeof(POS_S10_WRITE_PROTECTION_AREA_CMD) + i] = write_data[i];

  buffer_len = sizeof(POS_S10_WRITE_PROTECTION_AREA_CMD) + write_len;
  if (!SCardS10CardTransmit(hCard, buffer_len, buffer, sizeof(buffer), buffer, status))
      return FALSE;

  return TRUE;
}

/* SpringCardS10API/POS_S10_UPDATE_SECURITY_AREA_CMD
 *
 * NAME
 *   POS_S10_UPDATE_SECURITY_AREA_CMD
 *
 * DESCRIPTION
 *   S10 update security APDU command
 *   For GemCore POS Serial V1.5X
 *
 **/
static BYTE POS_S10_UPDATE_SECURITY_AREA_CMD[] =
{
  0x00, 0xD0, 0xC0, 0x00 /* Address */, 0x00 /* Update length*/ 	
};

/* SpringCardS10API/PosS10UpdateSecurityArea
 *
 * NAME
 *   PosS10UpdateSecurityArea
 *
 * DESCRIPTION
 *   S10 update security memory function using APDU exchange
 *   For GemCore POS Serial V1.5X
 *
 * INPUTS
 *   SCARDHANDLE hCard : handle of the card
 *   BYTE *update_data : data to update
 *   BYTE addr         : first address to update
 *   BYTE update_len   : length of data to update
 *   BYTE status[2]    : 2-byte buffer to receive status code
 *
 * RETURNS
 *   SCARD_S_SUCCESS   : success, data updated
 *   Other code if internal or communication error has occured.  
 *
 * NOTES
 *   This function must be used with a S10 card containing a security memory.
 *   This function will only be executed if a PSC has been successfully verified before.
 *   Maximum length of data to write is 4 bytes.
 *   Address lies between 0 and 3.
 *   For more information, read the datasheet of the S10 card.
 * 
 * SEE ALSO
 *   LiteProS10UpdateSecurityArea
 *   SCardS10UpdateSecurityArea
 *
 **/
LONG PosS10UpdateSecurityArea(SCARDHANDLE hCard, BYTE *update_data, BYTE addr, BYTE update_len, BYTE status[2])
{ 
  int i;
  BYTE buffer_len = 0;
  BYTE buffer[MAX_BYTE_TO_READ + 7];

  if (update_len > 4) 
    return SCARD_E_INVALID_PARAMETER;

  if (update_data == NULL) 
    return SCARD_E_INVALID_PARAMETER;

  if (! ((addr >= 0x00) && (addr <= 0x03)))
    return SCARD_E_INVALID_PARAMETER;

  POS_S10_UPDATE_SECURITY_AREA_CMD[3] = addr;
  POS_S10_UPDATE_SECURITY_AREA_CMD[4] = update_len;

  for (i=0 ; i < sizeof(POS_S10_UPDATE_SECURITY_AREA_CMD) ; i++)
	  buffer[i] = POS_S10_UPDATE_SECURITY_AREA_CMD[i];
  for (i=0 ; i < update_len ; i++)
	  buffer[sizeof(POS_S10_UPDATE_SECURITY_AREA_CMD) + i] = update_data[i];

  buffer_len = sizeof(POS_S10_UPDATE_SECURITY_AREA_CMD) + update_len;
  if (!SCardS10CardTransmit(hCard, buffer_len, buffer, sizeof(buffer), buffer, status))
      return FALSE;

  return TRUE;
}

/* SpringCardS10API/POS_S10_PSC_VERIFICATION_CMD
 *
 * NAME
 *   POS_S10_PSC_VERIFICATION_CMD
 *
 * DESCRIPTION
 *   S10 PSC verification APDU command
 *   For GemCore POS Serial V1.5X
 *
 **/
static BYTE POS_S10_PSC_VERIFICATION_CMD[] =
{
  0x00, 0x20, 0x00, 0x00, 0x03, 0x00 /* Code 3 */, 0x00 /* Code 2*/ , 0x00 /* Code 1*/	
};

/* SpringCardS10API/PosS10PSCVerification
 *
 * NAME
 *   PosS10PSCVerification
 *
 * DESCRIPTION
 *   S10 verification function of the Programmable Security Code (PSC) using APDU exchange
 *   For GemCore POS Serial V1.5X
 *
 * INPUTS
 *   SCARDHANDLE hCard : handle of the card
 *   BYTE code3        : byte 3 of Programmable Security Code
 *   BYTE code2        : byte 2 of Programmable Security Code
 *   BYTE code1        : byte 1 of Programmable Security Code
 *   BYTE status[2]    : 2-byte buffer to receive status code
 *
 * RETURNS
 *   SCARD_S_SUCCESS   : PSC verification succeeded
 *   Other code if internal or communication error has occured.  
 *
 * NOTES
 *   This function must be used with a S10 card containing a Programmable Security Code.
 *   For more information, read the datasheet of the S10 card.
 * 
 * SEE ALSO
 *   LiteProS10PSCVerification
 *   SCardS10PSCVerification
 *
 **/
LONG PosS10PSCVerification(SCARDHANDLE hCard, BYTE code3, BYTE code2, BYTE code1, BYTE status[2])
{ 
  LONG SCardLastError;
  BYTE buffer[MAX_BYTE_TO_READ + 2];

  POS_S10_PSC_VERIFICATION_CMD[5] = code3;
  POS_S10_PSC_VERIFICATION_CMD[6] = code2;
  POS_S10_PSC_VERIFICATION_CMD[7] = code1;

  SCardLastError = SCardS10CardTransmit(hCard, sizeof(POS_S10_PSC_VERIFICATION_CMD), POS_S10_PSC_VERIFICATION_CMD, sizeof(buffer), buffer, status);
  if (SCardLastError != SCARD_S_SUCCESS)
    return SCardLastError; 

  return SCARD_S_SUCCESS;
}